import glob
import os
import sys
import requests
import wave
import math

def normalizePath(path):
  return os.sep.join([os.getcwd(),path])


def wav2msu(filename,romname):
  trackinfo = filename.split(os.sep)[-1].split('.')[0].split('-')
  print(f"Converting track #{trackinfo[1]} to msu..")
  input  = wave.open(filename, 'rb')
  params = input.getparams()
  frames = input.readframes(params.nframes)
  input.close()
  if params.nchannels != 2 or params.sampwidth != 2 or params.framerate != 44100:
    print(f"Invalid file: {filename} with {params.nchannels} channels, bit depth of {params.sampwidth*8} and a sample rate of {params.framerate}")
  else:
    with open(normalizePath(f"msu\{romname}-{trackinfo[1]}.pcm"), 'wb') as f:
      f.write("MSU1".encode("utf-8"))
      f.write(int(trackinfo[2]).to_bytes(4, byteorder='little'))
      f.write(frames)


def patchIPS(patch, inprefix):
  print(f"Attempting to patch {patch} onto msu\{inprefix}.sfc ...", end="", flush=True)
  with open(patch, 'rb') as f:
    of = open(normalizePath(f"msu\{inprefix}.sfc"),'r+b')
    bytes = f.read(5)
    if bytes != b"PATCH":
      of.close()
      sys.exit("Not a valid IPS file")
    # Now check each record
  
    while(True):
      offset = int.from_bytes(f.read(3),"big")
      if offset == 0x454F46:
        print("Success!")
        break
      size   = int.from_bytes(f.read(2),"big")
      if size == 0:
        #RLE Encoded
        rle_size  = int.from_bytes(f.read(2),"big")
        rle_write = f.read(1)
        #print(f"At offset {hex(offset)} overwrite {rle_size} bytes with {hex(int.from_bytes(rle_write,'big'))}:")
        of.seek(offset)
        for x in range(rle_size):
          of.write(rle_write)
      else:
        #Standard Record
        record = f.read(size)
        #print(fr"At offset {hex(offset)} overwrite {size} bytes:")
        of.seek(offset)
        of.write(record)
    of.close()


if os.name == 'nt':
  os.system('cls')
else:
  os.system('clear')
  print("Running this on a non-Windows OS will require you to have flac and opusdec in your path to work.")

print("Choose which music track set you wish to download. Once they are processed, the full track set will take up about 3 GB of storage regardless of which one you download.")
print("1) Download lossless FLAC source tracks (Approx. 1.5 GB)")
print("2) Download high bitrate opus source tracks (Approx. 500 MB)")
print("3) Download low bitrate opus source tracks (Approx. 210 MB)")
print("4) Do not download or convert any tracks, only patch the rom.")
trackset = 0

while trackset not in ["1","2","3","4"]:
  trackset = input(">")
  
extension = "opus"
folder = "opus"
if trackset == "1":
  extension = "flac"
  folder = "flac"
elif trackset == "3":
  folder = "opuslow"
elif trackset == "4":
  extension = None
  folder = None

if extension:  
  try:
    os.mkdir(normalizePath("tracks"))
  except FileExistsError:
    pass

try:
  os.mkdir(normalizePath("optional"))
except FileExistsError:
  pass
  
if len(sys.argv) == 1:
  inromname = input("Please input your FF6 T-Edition rom file location (You may also drag the rom onto this window) >")
  
elif len(sys.argv) == 2:
  os.chdir(os.path.split(sys.argv[0])[0])
  inromname = sys.argv[1]
romname = ".".join(os.path.split(inromname)[1].split(".")[:-1])

inromname = inromname.lstrip('"').rstrip('"')
if not os.path.exists(inromname):
  sys.exit("Invalid path to ROM!")

response = ""
apply_optional = False
while response.lower() != "y" and response.lower() != "n":
  response = input('Did you want to apply any ips patches stored in the "optional" folder [Y/N]?')

if response.lower() == "y":
  apply_optional = True

#Do downloads here
print("Downloading tracklist, patch and utilities")
if extension:  
  url = "https://bk0.ovh/msu/tracklist.txt"
  tracklist = requests.get(url).content.decode('utf-8')

  if extension == "flac" and os.name == 'nt':
    url = "https://bk0.ovh/msu/flac.exe"
    r = requests.get(url)
    with open(normalizePath(f"flac.exe"), 'wb') as f:
      f.write(r.content)
  elif extension == "opus" and os.name == 'nt':
    url = "https://bk0.ovh/msu/opusdec.exe"
    r = requests.get(url)
    with open(normalizePath(f"opusdec.exe"), 'wb') as f:
      f.write(r.content)

url = "https://bk0.ovh/msu/msu.ips"
r = requests.get(url)
with open(normalizePath(f"msu.ips"), 'wb') as f:
  f.write(r.content)


try:
  os.mkdir(normalizePath("msu"))
except FileExistsError:
  pass


if extension:
  for track in tracklist.split('\r\n'):
    trackinfo = track.split("-")
    if not os.path.exists(normalizePath(f"tracks\{track}.{extension}")) and not os.path.exists(normalizePath(f"msu\{romname}-{trackinfo[1]}.pcm")):
      downloaded = 0.0
      percentage = 0.0
      headers = {}
      resume = 0
      fileExists = False
      if os.path.exists(normalizePath(f"tracks\{track}.{extension}.part")):
        fileExists = True
      with open(normalizePath(f"tracks\{track}.{extension}.part"), 'ab') as f:
        if fileExists:
          resume = f.tell()
          headers['Range'] = f"bytes={resume}-"
          print(f"Resuming {track}.{extension}")
        else:
          print(f"Downloading {track}.{extension}")
        url = f"https://bk0.ovh/msu/{folder}/{track}.{extension}"
        r = requests.get(url,headers=headers,stream=True)
        size = int(r.headers['content-length']) + resume
        downloaded = resume
        for chunk in r.iter_content(chunk_size=8192):
          downloaded = downloaded + len(chunk)
          if math.floor(downloaded / size *100) != math.floor(percentage*100):
            print(f"Downloaded {math.floor(percentage*100)}%      ", end='\r', flush=True)
          percentage = downloaded / size
          f.write(chunk)
      os.rename(normalizePath(f"tracks\{track}.{extension}.part"),normalizePath(f"tracks\{track}.{extension}"))
        

if extension:
  for filename in glob.glob(normalizePath('tracks\*.opus')):
    trackinfo = filename.split(os.sep)[-1].split('.')[0].split('-')
    outputTrack = normalizePath(f"tracks\{romname}-{trackinfo[1]}-{trackinfo[2]}.wav")
    if not os.path.exists(outputTrack):
      os.system(f'opusdec --force-wav "{filename}" "{outputTrack}"')
  for filename in glob.glob(normalizePath('tracks\*.flac')):
    trackinfo = filename.split(os.sep)[-1].split('.')[0].split('-')
    outputTrack = normalizePath(f"tracks\{romname}-{trackinfo[1]}-{trackinfo[2]}.wav")
    if not os.path.exists(outputTrack):
      os.system(f'flac -d "{filename}" -o "{outputTrack}"')
  for filename in glob.glob(normalizePath('tracks\*.wav')):
    wav2msu(filename,romname)
    os.remove(filename)
  print(os.linesep,os.linesep,os.linesep,os.linesep)  #make it clear that the following messages are not just processing output so maybe people will read them


with open(f'msu\{romname}.msu','wb') as f:
  f.flush()

#Copy our source rom to our destination before patching
sourcerom = open(inromname,'rb')
destrom = open(normalizePath(f"msu\{romname}.sfc"),'wb')
oldbin = sourcerom.read()
destrom.write(oldbin)
sourcerom.close()
destrom.close()

if apply_optional:
  for filename in glob.glob(normalizePath('optional\*.ips')):
    patchIPS(filename,romname)

patchIPS("msu.ips",romname)

if glob.glob(normalizePath('tracks\*.opus')) or glob.glob(normalizePath('tracks\*.flac')) or glob.glob(normalizePath('tracks\*.wav')):
  print('Note that tracks will not be re-downloaded as long as the pcm file exists in the "msu" directory\r\nIt should be unnecessary to keep them.')
  response = ""
  while response.lower() != "y" and response.lower() != "n":
    response = input("Did you want to delete all of the downloaded tracks [Y/N]?")

  if response.lower() == "y":
    for filename in glob.glob(normalizePath('tracks\*.opus')):
      os.remove(filename)
    for filename in glob.glob(normalizePath('tracks\*.flac')):
      os.remove(filename)
    # The WAV files should already be gone, but just in case
    for filename in glob.glob(normalizePath('tracks\*.wav')):
      os.remove(filename)

print('Operation complete! Patched rom and tracks should be located in the "msu" directory!')
os.system("pause")